
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

enum {
	// _xinfreq = 20_000_000,
	_xtlfreq = 20_000_000,
	_clkfreq = 4_000_000,
	DOWNLOAD_BAUD = 230_400,
	DEBUG_BAUD = DOWNLOAD_BAUD,

	MOSI = 32,
	CLK = 33,
	CS = 34,
	MISO = 35,
};


/*
 *      ____     Mode 0 : CPHA=0  CPOL=0       ____
 *   CS     |_________________________________|
 *             _   _   _   _   _   _   _   _
 *  CLK ______| |_| |_| |_| |_| |_| |_| |_| |______
 *           ___ ___ ___ ___ ___ ___ ___ ___
 * MOSI ----X___X___X___X___X___X___X___X___XXX----
 * MISO       7   6   5   4   3   2   1   0
 * 
 * 
 *      ____     Mode 3 : CPHA=1  CPOL=1       ____
 *   CS     |_________________________________|
 *      ______   _   _   _   _   _   _   _   ______
 *  CLK       |_| |_| |_| |_| |_| |_| |_| |_|
 *             ___ ___ ___ ___ ___ ___ ___ ___
 * MOSI ----XXX___X___X___X___X___X___X___X___X----
 * MISO         7   6   5   4   3   2   1   0
 * 
 * 
 * As a minimum, SPI requires just half clock cycle from CS low to first clock
 *      ____ <------>
 *   CS     |________________________ ......
 *                   _______
 *  CLK ____________|       |_______| ......
 * 
 */


static void  spi_shift_out( uint8_t *addr )
{
    __asm volatile {  // no optimising and enforces Fcache use - Needed to free up the FIFO
		rdfast	#0, addr    // setup the hubRAM FIFO hardware for sequencial writes
		modz	15   wz    // set Z flag, mask matching event for SETPAT
		neg	pa, #CLK & 0x20   wc    // INA/INB port select for SETPAT
		setpat	mask1, mask1    // Setup event:  CLK high and CS high
		setse1	#2<<6 | CS    // Setup event:  1=rising-edge, 2=falling-edge, 3=both-edges, 4=low-level, 6=high-level

//		rflong	pa    // read two bytes from FIFO (hubRAM)
//		rev	pa    // 32-bit endian bit-swap to suit big-endian shifted from lsb of register
//		movbyts	pa, #0x1b    // longword endian byte-swap to suit big-endian shifted from lsb of register
		rfword	pa    // read two bytes from FIFO (hubRAM)
		rev	pa    // 32-bit endian bit-swap
		rol	pa, #8
.retry
		dirl	#MISO    // disable/reset SPI RX smartpin
		wypin	pa, #MISO
		dirh	#MISO    // enable SPI RX smartpin
.wait
		jpat	#.retry    // CLK toggled while CS still high, treat as a glitch
		jnse1	#.wait    // wait while CS is high

		setse1	#1<<6 | CS    // Setup event:  1=rising-edge, 2=falling-edge, 3=both-edges, 4=low-level, 6=high-level
		rol	pa, #8
		wypin	pa, #MISO

		rep	@.rend, #0    // loops forever, 12 ticks per loop
		jse1	#.rend    // break loop when CS high
		testp	#MISO   wc   // check for empty buffer
	if_c	rfbyte	pa    // read byte from FIFO (hubRAM)
	if_c	rev	pa    // 32-bit endian bit-swap
	if_c	rol	pa, #8
	if_c	wypin	pa, #MISO
.rend
		ret

mask1		long	(1<<(CS & 0x1f)) | (1<<(CLK & 0x1f))   // CS/CLK pin mask for SETPAT
    }  // Exiting of Fcache/Cogexec triggers an implicit RDFAST - flushing remaining FIFO content to hubRAM
}



static void  filldat( uint8_t *addr, size_t len )
{
    for( size_t i = 0; i < len; i++ )    // bytes
        addr[i] = i;
}


void main(void)
{
    uint32_t  m, x, ret;
    uint8_t  txdata[500];

    _wrpin(CS, P_HIGH_15K | P_LOW_FLOAT | P_SYNC_IO | P_SCHMITT_A);  // do this first
    _drvh(CS);  // 15 kR pull-up resistor

    _waitms(200);
    printf( "\n   clkfreq = %d   clkmode = 0x%x\n", _clockfreq(), _clockmode() );

    _wrpin(CLK, P_SCHMITT_A);  // single ended

    m = P_SYNC_TX | P_SCHMITT_A | P_OE;  // single ended
    m |= ((CLK-MISO) & 0b111 | 0b0000 ) << 24;  // smartB (clock pin) input select
    x = 0b0_00000 | (8-1);  // continuous/stop-start buffer mode and 8-bit words
    _pinstart(MISO, m, x, 0);

//    memset( addr, 0, sizeof(txdata) );
    filldat( txdata, sizeof(txdata) );
    x = 0;

    while(1)
    {
        txdata[0] = x;
        ret = spi_shift_out( txdata, sizeof(txdata) );
        printf( " BlockID = %d    Byte Count = %d\n", x, ret );
        x = (x + 1) % 20;
    }
}
